// Copyright by Yves Goergen
// Website: http://beta.unclassified.de/
// This file is licensed under the terms of the GNU General Public License (GPL).

using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Threading;
using System.Diagnostics;
using System.Drawing;

namespace Unclassified.Automation
{
	public class Window
	{
		private delegate bool EnumWindowsDelegate(IntPtr handle, int lParam);

		[DllImport("user32")]
		private static extern int EnumWindows(EnumWindowsDelegate cb, int lparam);
		[DllImport("user32")]
		private static extern int GetWindowThreadProcessId(IntPtr hWnd, out int processId);

		private IntPtr handle = IntPtr.Zero;
		private WindowProperties searchProperties;

		/// <summary>
		/// Creates a new Window object finding an existing window by its handle.
		/// </summary>
		/// <param name="hWnd">Window handle to match</param>
		public Window(IntPtr hWnd)
		{
			handle = hWnd;
		}

		/// <summary>
		/// Creates a new Window object finding an existing window by search properties.
		/// </summary>
		/// <param name="properties"></param>
		public Window(WindowProperties properties)
		{
			searchProperties = properties;

			Find();
		}

		public IntPtr Handle
		{
			get
			{
				return handle;
			}
		}

		public int ProcessId
		{
			get
			{
				int processId;
				GetWindowThreadProcessId(handle, out processId);
				return processId;
			}
		}

		public Rectangle Rectangle
		{
			get
			{
				Rectangle rect;
				WinApi.GetWindowRect(handle, out rect);
				return rect;
			}
		}

		public uint Style
		{
			get
			{
				return WinApi.GetWindowLong(handle, WinApi.GWL_STYLE);
			}
		}

		public uint ExStyle
		{
			get
			{
				return WinApi.GetWindowLong(handle, WinApi.GWL_EXSTYLE);
			}
		}

		private void Find()
		{
			EnumWindows(EnumWindowsProcedure, 0);
			handle = searchProperties.Handle;
			if (handle == IntPtr.Zero) throw new WindowNotFoundException("");
		}

		/// <summary>
		/// Compares each found window with the search criteria.
		/// </summary>
		/// <param name="handle"></param>
		/// <param name="lParam"></param>
		/// <returns>true if the window does not match and the next window should be tested. false if the window matches</returns>
		private bool EnumWindowsProcedure(IntPtr handle, int lParam)
		{
			// Test each criteria
			if (searchProperties.ClassName != null && searchProperties.ClassName != GetClassName(handle)) return true;
			if (searchProperties.ProcessId != 0 && searchProperties.ProcessId != GetProcessId(handle)) return true;
			if (searchProperties.Visible != null && searchProperties.Visible != WinApi.IsWindowVisible(handle)) return true;

			// All required properties match, stop the search and save the window handle
			searchProperties.Handle = handle;
			return false;
		}

		/// <summary>
		/// Determines the class name of a window by its handle.
		/// </summary>
		/// <param name="handle"></param>
		/// <returns></returns>
		public static string GetClassName(IntPtr handle)
		{
			StringBuilder className = new StringBuilder(256);
			WinApi.GetClassName(handle, className, 256);
			return className.ToString();
		}

		/// <summary>
		/// Gets the process ID of the window.
		/// </summary>
		/// <returns></returns>
		public static int GetProcessId(IntPtr handle)
		{
			int processId;
			GetWindowThreadProcessId(handle, out processId);
			return processId;
		}

		public static bool TryFind(WindowProperties prop, out Window window)
		{
			try
			{
				window = new Window(prop);
				return true;
			}
			catch (WindowNotFoundException)
			{
				window = null;
				return false;
			}
		}
	}
}
